/*Open assistive arm for meals code
 *  This firmware is used to calibrate the device before using by someone.
 *  version 1 08/09/2020
 *  edited by Georges Tawil 
 *  contact : georges.tawil@apf.asso.fr
 * 
 * 
 * 
 */
// Import libraries:
//    - dynamixel shield library can be get from here : https://emanual.robotis.com/docs/en/parts/interface/dynamixel_shield/
#include <Dynamixel2Arduino.h>

// Define DXL Motors communication
#define       DXL_SERIAL   Serial
const uint8_t DXL_DIR_PIN = 2;              // DYNAMIXEL Shield DIR PIN
const float   DXL_PROTOCOL_VERSION = 2.0; 

typedef struct{
  uint8_t   id                      ;
  bool      availability            ;
  int       max_speed               ;   // Profile Velocity
  float     calibrationParameter[3] ;
  bool      enable                  ;
  float     current_angle           ;
  } Actuator;
  
Dynamixel2Arduino dxl(DXL_SERIAL, DXL_DIR_PIN);

//*********************************Interrupts****************************************************
#define interruptPin 18 // defining pin 18 as an interrupt pin
unsigned long t_high=0, t_low=0, t=0, t_but = 0;
bool but_rise = false,  but_fall = false;

//time intervals of pressing the button (milliseconds)
#define BETWEEN_0_200       0
#define BETWEEN_200_700     1
#define BETWEEN_700_4000    2

int  but_status = BETWEEN_0_200;          // Init button status
bool but_last  = false, but_new = false;

bool action_required = false, action_done = false;
//***********************************************************************************************

#define ALL      0

const float REAL_POSITION[6][2] {
  {0.0, -90.0},    
  {0.0, 0.0},        
  {0.0, 0.0},
  {0.0, -90.0},
  {0.0, -90.0},
  {0.0, 90.0}
};

// Create a table for describing all actuators of the robot arm
Actuator actuator_tab[6];

float savedPosition[6];
int saved_position_counter = 0;
int calibration_step=0;

float cmd[6];

float pos1[6]={-45.0, -45.0, -45.0, 0.0, 0.0, 0.0};
float pos2[6]={20.0, 20.0, -45.0, -20.0, -20.0, 90.0};
float pos3[6]={0.0, 0.0, 0.0, 0.0, 0.0, 0.0};

void setup() {
  // Set Port baudrate to 57600bps. This has to match with DYNAMIXEL baudrate.
  dxl.begin(57600);
  // Set Port Protocol Version. This has to match with DYNAMIXEL protocol version.
  dxl.setPortProtocolVersion(DXL_PROTOCOL_VERSION);

  // Interrupts Init
  pinMode(interruptPin, INPUT);  // pin definition as input
  attachInterrupt(digitalPinToInterrupt(interruptPin), button_change, CHANGE);  // call function button_change() in the state of button was changed

  
  //                {id , availability  , max_speed  ,  calibrationParameter  ,  enable , current_angle };
  actuator_tab[0] = {1  ,     false     ,     30     ,    {0.0, 0.0, 0.0}     ,   false ,        0     };
  actuator_tab[1] = {2  ,     false     ,     30     ,    {0.0, 0.0, 0.0}     ,   false ,        0     };
  actuator_tab[2] = {3  ,     false     ,     30     ,    {0.0, 0.0, 0.0}     ,   false ,        0     };
  actuator_tab[3] = {4  ,     false     ,     30     ,    {0.0, 0.0, 0.0}     ,   false ,        0     };
  actuator_tab[4] = {5  ,     false     ,     30     ,    {0.0, 0.0, 0.0}     ,   false ,        0     };
  actuator_tab[5] = {6  ,     false     ,     30     ,    {0.0, 0.0, 0.0}     ,   false ,        0     };


  if (actuator_availability_checking()){
    Serial.println();
    Serial.println("***************      Actuators are connected and work     ********************");
    Serial.println();
  }
}

void loop() {
  read_data_motor();
  //print_data_motor();
  // Check button status
  check_times();
  if (action_required && !action_done){
    do_action();
    action_done = true;
  }
}

/* Actuator availability checking
 *  This function checks that every actuators are wiring and available to be controled, it returns an boolean:
 *    -false : there is one or more actuator is are not available
 *    -true : all actuators are available, they can be used
 *  la fonction permet de vérifier que tous les actionneurs sont connectés et disponible pour le controle
 */
bool actuator_availability_checking(){
  bool availability = true;
  for (int i=0; i<6; i++){
    actuator_tab[i].availability = dxl.ping(actuator_tab[i].id);
    if (actuator_tab[i].availability){
      dxl.torqueOff(actuator_tab[i].id);
      dxl.setOperatingMode(actuator_tab[i].id, OP_EXTENDED_POSITION);
      dxl.writeControlTableItem(PROFILE_VELOCITY,     actuator_tab[i].id, actuator_tab[i].max_speed);
      dxl.writeControlTableItem(PROFILE_ACCELERATION, actuator_tab[i].id, actuator_tab[i].max_speed/2);
      dxl.torqueOn(actuator_tab[i].id);
    }
    else{
      availability = false;
    }
  }
  if (availability){
    Serial.println();
    Serial.println("***************      Actuators are connected and work     ********************");
    Serial.println();
  }
  return availability;
}

/* Read actuator datas
 *  Read every data of actuators and push them in actuators table with updated data
 */
void read_data_motor(){
  for (int i=0; i<6; i++){
    actuator_tab[i].current_angle  = dxl.getPresentPosition(actuator_tab[i].id, UNIT_DEGREE);   // Read angle 
  }
}

void print_data_motor(){
  Serial.println();
  Serial.print("Present Position:");
  for (int i=0; i<6; i++){
    Serial.print("  q");  Serial.print(actuator_tab[i].id); Serial.print(": "); Serial.print(actuator_tab[i].current_angle);
  }
  Serial.println();
}

void torque_on(int id){
  switch (id){
    case ALL:
      for (int i=0; i<6; i++){
        if (actuator_tab[i].availability){
          dxl.torqueOn(actuator_tab[i].id);
        }
      }
      break;
    default:
      dxl.torqueOn(actuator_tab[id-1].id);
  }
}

void torque_off(int id){
  switch (id){
    case ALL:
      for (int i=0; i<6; i++){
        if (actuator_tab[i].availability){
          dxl.torqueOff(actuator_tab[i].id);
        }
      }
      break;
    default:
      dxl.torqueOff(actuator_tab[id-1].id);
  }
}

void button_change(){
  bool but_new = digitalRead(interruptPin);
  //Serial.print("but_last = "); Serial.print(but_last); Serial.print(" but_new = "); Serial.println(but_new);
  if (but_new > but_last) button_rise();
  else if (but_new < but_last) button_fall();
  but_last = but_new;
}

void button_rise(){
  but_rise   = true;
  but_fall   = false;
  t = millis();
}

void button_fall(){
  but_fall   = true;
  but_rise   = false;
  t = millis();
}

void check_times(){
  long int dt;
  dt = millis() - t;
  if(but_rise){
    action_required = false;
    t_high = dt;
    t_low  = 0;
  }
  else if(but_fall){
    t_low  = dt;
    t_high = 0;
  }

  //Serial.println(dt);

  if      (t_high > 700)                    { torque_off(ALL); }
  
  if      (t_high > 0   && t_high < 200  )  { but_status = BETWEEN_0_200;    }
  else if (t_high > 200 && t_high < 1000  )  { but_status = BETWEEN_200_700; action_done = false; }
  else if (t_high > 2000 && t_high < 4000 )  { but_status = BETWEEN_700_4000; action_done = false;}
  else if (t_high > 4000 ) { but_status = BETWEEN_0_200; }
  if      (t_low > 200 && !action_done)     { torque_on(ALL); action_required = true;}
}

void print_saved_position(int joint_id, float angle){
  Serial.println();
  Serial.println("***************      Position saved     ********************");
  Serial.print("            Joint :   "); Serial.print(joint_id); Serial.print(",        readed position :   ");Serial.println(angle);
  Serial.println();
}

void do_action(){
  switch(but_status){
    case BETWEEN_200_700:
    if (calibration_step<3){
      switch(calibration_step){
        case 0:
        if (saved_position_counter<2){
          print_saved_position(actuator_tab[0].id, actuator_tab[0].current_angle);
          savedPosition[saved_position_counter]=actuator_tab[0].current_angle;
          print_saved_position(actuator_tab[3].id, actuator_tab[3].current_angle);
          savedPosition[saved_position_counter+3]=actuator_tab[3].current_angle;
          saved_position_counter++;
        }
        if (saved_position_counter==2){
          actuator_tab[0].calibrationParameter[0]=(savedPosition[1]-savedPosition[0])/-90;
          actuator_tab[0].calibrationParameter[1]=savedPosition[0];
          actuator_tab[3].calibrationParameter[0]=(savedPosition[4]-savedPosition[3])/-90;
          actuator_tab[3].calibrationParameter[1]=savedPosition[3];
          calibration_step++;
          saved_position_counter=0;
        }
        break;

        case 1:
        if (saved_position_counter<3){
          print_saved_position(actuator_tab[1].id, actuator_tab[1].current_angle);
          savedPosition[saved_position_counter]=actuator_tab[1].current_angle;
          print_saved_position(actuator_tab[2].id, actuator_tab[2].current_angle);
          savedPosition[saved_position_counter+3]=actuator_tab[2].current_angle;
          saved_position_counter++;
        }
        if (saved_position_counter==3){
          actuator_tab[1].calibrationParameter[2]=savedPosition[2];
          actuator_tab[1].calibrationParameter[1]=(savedPosition[1]-actuator_tab[1].calibrationParameter[2])/90;
          actuator_tab[1].calibrationParameter[0]=(savedPosition[0]-actuator_tab[1].calibrationParameter[2])/90;
          actuator_tab[2].calibrationParameter[2]=savedPosition[5];
          actuator_tab[2].calibrationParameter[1]=(savedPosition[3]-actuator_tab[2].calibrationParameter[2])/90;
          actuator_tab[2].calibrationParameter[0]=(savedPosition[4]-actuator_tab[2].calibrationParameter[2])/90;
          calibration_step++;
          saved_position_counter=0;
        }
        break;

        case 2:
        if (saved_position_counter<2){
          print_saved_position(actuator_tab[4].id, actuator_tab[4].current_angle);
          savedPosition[saved_position_counter]=actuator_tab[4].current_angle;
          print_saved_position(actuator_tab[5].id, actuator_tab[5].current_angle);
          savedPosition[saved_position_counter+3]=actuator_tab[5].current_angle;
          saved_position_counter++;
        }
        if (saved_position_counter==2){
          actuator_tab[4].calibrationParameter[0]=(savedPosition[1]-savedPosition[0])/-90;
          actuator_tab[4].calibrationParameter[1]=savedPosition[0];
          actuator_tab[5].calibrationParameter[0]=(savedPosition[4]-savedPosition[3])/90;
          actuator_tab[5].calibrationParameter[1]=savedPosition[3];
          calibration_step++;
          saved_position_counter=0;
        }
        break;

        default:
        break;
      }
    }
    else{
      printMessage("calibration set up");
      for (int i=0; i<6; i++){
        Serial.println();
        Serial.print("            Joint : ");
        Serial.print(actuator_tab[i].id);
        Serial.print("           a : ");
        Serial.print(actuator_tab[i].calibrationParameter[0]);
        Serial.print("           b : ");
        Serial.print(actuator_tab[i].calibrationParameter[1]);
        Serial.print("           c : ");
        Serial.print(actuator_tab[i].calibrationParameter[2]);
        Serial.println();
      }
      test(); 
    }    
    break;
      
    case BETWEEN_700_4000:
      Serial.println();
      Serial.println("***************      Restart calibration from the first joint     ********************");
      Serial.println();
      //saved_position_counter=0;
      break;
      
    default:
    Serial.println("Nothing");
    print_data_motor();
    Serial.println("Nothing");
    break;
  }
}

void test(){
  delay(2000);
      Serial.println();
      Serial.println("***************      move     ********************");
      Serial.println();
      delay(2000);
      moveTo(pos3);
      delay(10000);
      moveTo(pos1);
      delay(10000);
      moveTo(pos2);
      delay(10000);
      while(1){}
}

void modelPosition2command(float modelPosition[6]){
  cmd[0]=modelPosition[0]*actuator_tab[0].calibrationParameter[0]+actuator_tab[0].calibrationParameter[1];
  cmd[1]=modelPosition[1]*actuator_tab[1].calibrationParameter[0]+modelPosition[2]*actuator_tab[1].calibrationParameter[1]+actuator_tab[1].calibrationParameter[2];
  cmd[2]=modelPosition[2]*actuator_tab[2].calibrationParameter[0]+modelPosition[1]*actuator_tab[2].calibrationParameter[1]+actuator_tab[2].calibrationParameter[2];
  cmd[3]=modelPosition[3]*actuator_tab[3].calibrationParameter[0]+actuator_tab[3].calibrationParameter[1];
  cmd[4]=modelPosition[4]*actuator_tab[4].calibrationParameter[0]+actuator_tab[4].calibrationParameter[1];
  cmd[5]=modelPosition[5]*actuator_tab[5].calibrationParameter[0]+actuator_tab[5].calibrationParameter[1];
}

void moveTo(float Q[6]){
  modelPosition2command(Q);
  for (int i=0; i<6; i++){
    dxl.setGoalPosition(actuator_tab[i].id, cmd[i], UNIT_DEGREE);
  }
}



void printMessage(const char *message){
  Serial.println();
  Serial.print("************************         ");
  Serial.print(message);
  Serial.println("         ************************");     
  Serial.println();
}
